#!/bin/bash
########################################################################
# build_libtorch.sh
#
# Build libtorch and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_02.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=libtorch
origpkgver=v1_5_1
pkgver=${origpkgver} # Append letters for local patches.
ssibuildshims_version=v2_04_01
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${pkgdotver}
pkgtarfile=${pkgtarname}.tar
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e2[01],c7} \
  --bq-blas={open,eigen} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  CET_USE_NINJA

    Use Ninja if available, or fall back to make. Defaults to 1.

  DEBUG_(CONFIG|BUILD)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

if [ "$blas" = eigen ] && ! item_in $cqual c7 e19; then
  ssi_die "BLAS implentation $blas not supported with $cqual"
fi

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics IN_SOURCE NO_TYPE TRACK_PROGRESS || exit # *C*

# Define versions of build-time-only products to be set up.
case ${blas} in
  open)
    libtorch_blas=OpenBLAS
    pybind11_default_version=v2_5_0a
    fftw_default_version=v3_3_8a
    ;;
  eigen)
    libtorch_blas=Eigen
    pybind11_default_version=v2_2_4
    eigen_default_version=v3_3_5
    fftw_default_version=v3_3_8a
    ;;
  *) ssi_die -I "accepted unknown blas=${blas}"
esac
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit

####################################

####################################
# Build-only dependencies. *C*

# Build tools.
${trace} setup_cmake --latest || exit
(( ${CET_USE_NINJA:-1} )) && \
  ${trace} setup_ninja --optional && \
  (( want_ninja = 1)) || true

# Python is required for the build's "setup" program but not otherwise
# with current configurations.
${trace} setup_python || exit

##################
# These packages are either header-only or expose neither dependency nor
# defined symbols. Default versions are determined above.
if [ "$libtorch_blas" = Eigen ]; then
  ${trace} setup_prod eigen -- -q +|| exit
fi

${trace} setup_prod fftw -- -q +prof && \
  ${trace} setup_prod pybind11 -- -q + || exit
##################
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  || exit

# Apply source tweaks
${trace} find \
  ${sourcedir}/${pkgtarname} \( -name 'FindOpenBLAS.cmake' \
  -o -name 'FindOpenMP.cmake' -o -name 'FindNumPy.cmake' \
  -o -name 'Findpybind11.cmake' \) -print -delete || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a cflg cxxflg compile_options

case ${cqual} in
  e19) cc=gcc; cxx=g++; cxxstd=17; cxxflg+=(-Wno-deprecated-copy);;
  e20) cc=gcc; cxx=g++; cxxstd=17; cxxflg+=(-Wno-deprecated-copy);;
  e21) cc=gcc; cxx=g++; cxxstd=20; cxxflg+=(-Wno-deprecated-copy);;
  c7) cc=clang; cxx=clang++; cxxstd=17;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac

cflg+=( \
  -Wno-deprecated-declarations \
  -Wno-parentheses \
)

cxxflg+=( \
  -Wno-deprecated-declarations \
  -Wno-parentheses \
  -Wno-redundant-move \
  -std=c++${cxxstd} \
  )

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}

# For ctest:
[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}/${pkgtarname}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

compile_options+=( \
  ${cflg[*]:+CFLAGS="${cflg[*]}"} \
  ${cxxflg[*]:+CXXFLAGS="${cxxflg[*]}"} \
  ${want_ninja:+CMAKE_GENERATOR=Ninja} \
  BLAS=$libtorch_blas \
  BUILD_BINARY=1 \
  BUILD_CUSTOM_PROTOBUF=0 \
  BUILD_TEST=0 \
  CC=${cc} \
  CXX=${cxx} \
  MAX_JOBS=${ncores} \
  REL_WITH_DEB_INFO=1 \
  USE_ASAN=0 \
  USE_CUDA=0 \
  USE_CUDNN=0 \
  USE_DISTRIBUTED=0 \
  USE_FBGEMM=0 \
  USE_FFMPEG=0 \
  USE_GFLAGS=0 \
  USE_GLOG=0 \
  USE_LEVELDB=0 \
  USE_LMDB=0 \
  USE_MKLDNN=0 \
  USE_MPI=0 \
  USE_NCCL=0 \
  USE_NNPACK=1 \
  USE_NUMA=0 \
  USE_NUMPY=0 \
  USE_NVRTC=0 \
  USE_OBSERVERS=1 \
  USE_OPENCL=0 \
  USE_OPENCV=0 \
  USE_OPENMP=1 \
  USE_PROF=0 \
  USE_QNNPACK=1 \
  USE_REDIS=0 \
  USE_ROCKSDB=0 \
  USE_ROCM=0 \
  USE_TBB=0 \
  USE_TENSORRT=0 \
  USE_XNNPACK=1 \
  USE_ZSTD=0 \
)

# Use our Eigen rather than building one (ours will be found over one
# under /usr due to CMAKE_PREFIX_PATH).
[ libtorch_blas = "Eigen" ] && \
  compile_options+=(USE_SYSTEM_EIGEN_INSTALL=1)

(( DEBUG_CONFIG )) || (( DEBUG_BUILD )) && compile_options+=(VERBOSE=1)

# Some CMake steering options.
compile_options+=( \
  CMAKE_NO_VERBOSE=$(( ! DEBUG_CONFIG )) \
  CMAKE_CXX_STANDARD=$cxxstd \
)

# Fix issues with older protobuf.
if ver_greater 3.12.3 $PROTOBUF_VERSION; then
  # Make it impossible to find this version's (broken)
  # protbuf-config.cmake so that we fall back on CMake's
  # FindProtobuf.cmake. This will honor the CMAKE definitions below for
  # Protobuf_* and do the rest correctly.
  CMAKE_PREFIX_PATH=$(dropit -p "${CMAKE_PREFIX_PATH}" \
    "${PROTOBUF_LIB}") || \
    ssi_die "unable to correct CMAKE_PREFIX_PATH for protobuf $PROTOBUF_VERSION"
  PATH=$(dropit "${PROTOBUF_FQ_DIR}")
  # Need to patch the libtorch build system so that we can send
  # arguments directly to CMake in addition to the usual method, which
  # is via a set of environment variables matching patterns or found in
  # an allowed list.
  ${trace} patch_source "${ssi_args[@]}" -P "${patchdir}" \
    -a "${sourcedir}/${pkgtarname}" \
    allow_extra_cmake_args.patch \
    || exit
  extra_cmake_args=(-- \
    -DProtobuf_INCLUDE_DIR=$PROTOBUF_INC \
    -DProtobuf_LIBRARY=$PROTOBUF_LIB/libprotobuf.so \
    -DProtobuf_LITE_LIBRARY=$PROTOBUF_LIB/libprotobuf-lite.so \
    -DProtobuf_PROTOC_LIBRARY=$PROTOBUF_LIB/libprotoc.so \
    -DProtobuf_PROTOC_EXECUTABLE=$PROTOBUF_FQ_DIR/bin/protoc \
    )
fi
##################

##################
# Build. *C*

# Some extra Python packages are required for building.
python_extra=(pyyaml)
# The typing module has been part of the Python Standard Library since
# 3.5. From 3.7 onward having it installed separately can cause
# issues. See
# e.g. https://fernandomc.com/posts/issues-with-typing-python-3/.
${trace} ver_greater 3.5.0 $PYTHON_VERSION && python_extra+=(typing)

py_pkgdir=${blddir}/python-packages
[[ -d "${py_pkgdir}" ]] && rm -rf "${py_pkgdir}"

mkdir "${py_pkgdir}" && \
  ${trace} pip install -t "${py_pkgdir}" "${python_extra[@]}" || exit
export PYTHONPATH=${py_pkgdir}${PYTHONPATH:+:${PYTHONPATH}}

${trace} env "${compile_options[@]}" python \
  ${sourcedir}/${pkgtarname}/tools/build_libtorch.py \
  "${extra_cmake_args[@]}" || \
  ssi_die "Unable to build ${package}"
##################

##################
# Test. *C*
##################

##################
# Install. *C*
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

# Where the build script installs everything...
install_tmp=${sourcedir}/${pkgtarname}/torch

# ...mixed in with some source we don't want.
${trace} find "$install_tmp/lib" -mindepth 1 -maxdepth 1 -type d \
  -exec rm -rf \{\} \; -prune

rsync_cmd=(rsync -lptgoD)
"${rsync_cmd[@]}" -r "${install_tmp}/"{include,lib,lib64,share} "${pkgdir}/" || \
  ssi_die "Unable to move temporary install files to ${pkgdir}/"

# Fix hard-wired path in one file:
perl -wapi\~ -e \
  'm&^\s*(?i)SET(?-i)\s*\(\s*ATEN_INCLUDE_DIR& and s&/.*tmp_install/include\b&\$ENV{LIBTORCH_INC}&' \
  "${pkgdir}/share/cmake/ATen/ATenConfig.cmake" && \
  rm -f "${pkgdir}/share/cmake/ATen/ATenConfig.cmake~" || \
  ssi_die "Unable to fix path in ATenConfig.cmake"

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
